import torch.nn as nn
import torch
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
import torch.nn.functional as F
# let's define a simple model that can deal with multimodal variable length sequence
class LFLSTM(nn.Module):

    def __init__(self, input_sizes, embed_size, hidden_sizes, fc1_size, output_size, dropout_rate, mode):
        super(LFLSTM, self).__init__()
        self.input_size = input_sizes
        self.hidden_size = hidden_sizes
        self.fc1_size = fc1_size
        self.output_size = output_size
        self.dropout_rate = dropout_rate
        self.mode = mode
        # defining modules - two layer bidirectional LSTM with layer norm in between
        if self.mode == 'text':
            self.embed = nn.Embedding(embed_size, input_sizes)
        self.trnn1 = nn.LSTM(input_sizes, hidden_sizes, bidirectional=True,num_layers=2)
        self.trnn2 = nn.LSTM(2 * hidden_sizes, hidden_sizes, bidirectional=True)
        if self.mode == 'visual':
            self.scale = 2
        else:
            self.scale = 4
        self.fc_t = nn.Linear(hidden_sizes * self.scale, fc1_size)
        self.bn_t = nn.BatchNorm1d(hidden_sizes * self.scale)
        self.fc2_t = nn.Linear(fc1_size, output_size)
        self.bn2_t = nn.BatchNorm1d(fc1_size)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout_rate)
        self.tlayer_norm = nn.LayerNorm((hidden_sizes * 2,))

    def extract_features(self, sequence, lengths, rnn1, rnn2, layer_norm):

        packed_sequence = pack_padded_sequence(sequence, lengths)
        packed_h1, (final_h1, _) = rnn1(packed_sequence)
        padded_h1, _ = pad_packed_sequence(packed_h1)
        normed_h1 = layer_norm(padded_h1)
        packed_normed_h1 = pack_padded_sequence(normed_h1, lengths)
        _, (final_h2, _) = rnn2(packed_normed_h1)
        return final_h1, final_h2

    def fusion(self, sentences, lengths):
        batch_size = lengths.size(0)
        if self.mode == 'text':
            sentences = self.embed(sentences)

        final_h1t, final_h2t = self.extract_features(sentences, lengths, self.trnn1, self.trnn2, self.tlayer_norm)
        if self.mode == 'visual':
            h_t = final_h1t.permute(1, 0, 2).contiguous().view(batch_size, -1)
        else:
            h_t = torch.cat((final_h1t, final_h2t)).permute(1, 0, 2).contiguous().view(batch_size, -1)
        return self.bn_t(self.relu(h_t))

    def forward(self, sentences, lengths):
        batch_size = lengths.size(0)
        h_t = self.fusion(sentences, lengths)
        h_t = self.fc_t(h_t)
        h_t = self.dropout(h_t)
        h_t = self.bn2_t(self.relu(h_t))
        o_t = self.fc2_t(h_t)
        return o_t


class LFLSTM_2(nn.Module):

    def __init__(self, input_sizes, embed_size, hidden_sizes, fc1_size, output_size, dropout_rate, mode):
        super(LFLSTM_2, self).__init__()
        self.input_size = input_sizes
        self.hidden_size = hidden_sizes
        self.fc1_size = fc1_size
        self.fc2_size = 32
        self.output_size = output_size
        self.dropout_rate = dropout_rate
        self.mode = mode
        # defining modules - two layer bidirectional LSTM with layer norm in between
        if self.mode == 'text':
            self.embed = nn.Embedding(embed_size, input_sizes)
        self.trnn1 = nn.LSTM(input_sizes, hidden_sizes, bidirectional=True, num_layers=2)
        self.scale = 2
        self.fc_t = nn.Linear(hidden_sizes * self.scale, fc1_size)
        self.bn_t = nn.BatchNorm1d(hidden_sizes * self.scale)
        self.fc2_t = nn.Linear(fc1_size, self.fc2_size)
        self.bn2_t = nn.BatchNorm1d(fc1_size)
        self.bn3_t = nn.BatchNorm1d(self.fc2_size)
        self.fc3_t = nn.Linear(self.fc2_size, output_size)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout_rate)
        self.tlayer_norm = nn.LayerNorm((hidden_sizes * 2,))

    def extract_features(self, sequence, lengths, rnn1):
        packed_sequence = pack_padded_sequence(sequence, lengths)
        packed_h1, (final_h1, _) = rnn1(packed_sequence)

        return final_h1

    def fusion(self, sentences, lengths):
        batch_size = lengths.size(0)
        # if self.mode == 'text':
        #     sentences = self.embed(sentences)

        final_h1t = self.extract_features(sentences, lengths, self.trnn1).view(2, 2, batch_size, self.hidden_size)[-1]
        if self.mode == 'visual':
            h_t = final_h1t.permute(1, 0, 2).contiguous().view(batch_size, -1)
        else:
            h_t = final_h1t.permute(1, 0, 2).contiguous().view(batch_size, -1)
        return self.bn_t(self.relu(h_t))

    def forward(self, sentences, lengths):
        h_t = self.fusion(sentences, lengths)
        h_t = self.fc_t(h_t)
        h_t = self.dropout(h_t)
        h_t = self.bn2_t(self.relu(h_t))
        h_t = self.fc2_t(h_t)
        h_t = self.bn3_t(F.relu(h_t))
        o_t = self.fc3_t(h_t)
        return o_t