import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import random
import numpy as np
import math
from torch.autograd import Variable
from util.layers import FC_basic_block, FC_basic_block_embed, GraphConvolution, \
    GraphConv_cheby, GraphAttentionLayer_MH


def get_fine_tuning_parameters(model, ft_begin_index):
    if ft_begin_index == 0:
        return model.parameters()

    ft_module_names = []
    for i in range(ft_begin_index, 5):
        ft_module_names.append('denseblock{}'.format(i))
        ft_module_names.append('transition{}'.format(i))
    ft_module_names.append('norm5')
    ft_module_names.append('classifier')

    parameters = []
    for k, v in model.named_parameters():
        for ft_module in ft_module_names:
            if ft_module in k:
                parameters.append({'params': v})
                break
        else:
            parameters.append({'params': v, 'lr': 0.0})

    return parameters


class _DenseLayer(nn.Sequential):

    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module('norm_1', nn.BatchNorm3d(num_input_features))
        self.add_module('relu_1', nn.ReLU(inplace=True))
        self.add_module('conv_1',
                        nn.Conv3d(
                            num_input_features,
                            bn_size * growth_rate,
                            kernel_size=1,
                            stride=1,
                            bias=False))
        self.add_module('norm_2', nn.BatchNorm3d(bn_size * growth_rate))
        self.add_module('relu_2', nn.ReLU(inplace=True))
        self.add_module('conv_2',
                        nn.Conv3d(
                            bn_size * growth_rate,
                            growth_rate,
                            kernel_size=3,
                            stride=1,
                            padding=1,
                            bias=False))
        self.drop_rate = drop_rate

        self.dp = nn.Dropout(self.drop_rate)

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            #print('using dp in densenet')
            new_features = self.dp(new_features)  # F.dropout(
        # new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):

    def __init__(self, num_layers, num_input_features, bn_size, growth_rate,
                 drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate,
                                growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _Transition(nn.Sequential):

    def __init__(self, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module('norm', nn.BatchNorm3d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=True))
        self.add_module('conv',
                        nn.Conv3d(
                            num_input_features,
                            num_output_features,
                            kernel_size=1,
                            stride=1,
                            bias=False))
        self.add_module('pool', nn.AvgPool3d(kernel_size=2, stride=2))


class DenseNet_l3(nn.Module):
    """Densenet-BC model class
    Args:
        growth_rate (int) - how many filters to add each layer (k in paper)
        block_config (list of 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
          (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
    """

    def __init__(self,
                 sample_size=64,
                 sample_duration=64,
                 inchannel=1,
                 growth_rate=32,
                 block_config=(6, 12, 24, 16),
                 num_init_features=32,
                 bn_size=4,
                 drop_rate=0,
                 num_classes=1000,
                 classes_config=(2),
                 model_k=32,
                 gc_num=2,
                 kernel_mul=1,
                 K=1,
                 edges=153,
                 in_fea_num=1):

        super(DenseNet_l3, self).__init__()

        self.sample_size = sample_size
        self.sample_duration = sample_duration
        self.num_classes = num_classes
        self.k = model_k
        self.gc_num = gc_num
        self.kernel_mul = kernel_mul
        self.K = K
        self.drop_rate = drop_rate
        self.edges = edges
        self.in_fea_num = in_fea_num

        # First convolution
        self.features = nn.Sequential(
            OrderedDict([
                ('conv0',
                 nn.Conv3d(
                     inchannel,
                     num_init_features,
                     kernel_size=3,
                     stride=1,
                     padding=1,
                     bias=False)),
                ('norm0', nn.BatchNorm3d(num_init_features)),
                ('relu0', nn.ReLU(inplace=True)),
                ('conv01', nn.Conv3d(num_init_features, num_init_features, kernel_size=3, stride=1,
                                     padding=1, bias=False)),
                ('norm01', nn.BatchNorm3d(num_init_features)),
                ('relu01', nn.ReLU(inplace=True))
            ]))

        # Each denseblock
        num_features = num_init_features
        multiple = 1
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate)
            self.features.add_module('denseblock%d' % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config):
                trans = _Transition(
                    num_input_features=num_features,
                    num_output_features=num_features // 2)
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = num_features // 2
                multiple = multiple * 2

        # Final batch norm
        self.features.add_module('norm5', nn.BatchNorm3d(num_features))
        last_duration = int(math.ceil(self.sample_duration / multiple))
        last_size = int(math.floor(self.sample_size / multiple))
        self.final_pool = nn.AvgPool3d(kernel_size=(last_duration, last_size, last_size))
        self.act = nn.ReLU(inplace=True)

        self.hidden = self.k

        # Linear layer
        if self.num_classes > 1:
            # ?
            self.classifier = [nn.Linear(num_features, classes_config[idx]) for idx in range(self.num_classes)]
            self.classifier = nn.ModuleList(self.classifier)
        # att_num = 0
        # for idx in range(self.num_classes - 1):
        #		att_num += classes_config[idx]
        # self.final_classifier = nn.Linear(num_features + att_num, classes_config[8])
        else:
            self.classifier = nn.Linear(num_features, classes_config[0])

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                m.weight = nn.init.kaiming_normal_(m.weight, mode='fan_out')
            elif isinstance(m, nn.BatchNorm3d) or isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x, adj=None, feature=False):

        features = self.features(x)
        out = self.act(features)

        out_features = self.final_pool(out).view(features.size(0), -1)

        if self.num_classes == 1:
            out = self.classifier(out_features)
            out = [out]
        else:
            out = []
            for idx in range(self.num_classes):
                out.append(self.classifier[idx](out_features))
        # att_cat = torch.cat(out, dim=1)
        # out.append(self.final_classifier(torch.cat([out_features, att_cat], dim=1)))
        # out.append(self.final_classifier(torch.cat([out_features, att_cat], dim=1)))

        if feature:
            return out, out_features
        else:

            return out



