seed = 123
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
import math
import numpy as np

class my_sparse_mm(torch.autograd.Function):
    """
    Implementation of a new autograd function for sparse variables,
    called "my_sparse_mm", by subclassing torch.autograd.Function
    and implementing the forward and backward passes.
    """

    def forward(self, W, x):  # W is SPARSE
        self.save_for_backward(W, x)
        y = torch.mm(W, x)
        return y

    def backward(self, grad_output):
        W, x = self.saved_tensors
        grad_input = grad_output.clone()
        grad_input_dL_dW = torch.mm(grad_input, x.t())
        grad_input_dL_dx = torch.mm(W.t(), grad_input)
        return grad_input_dL_dW, grad_input_dL_dx

class FC_basic_block(nn.Module):
    def __init__(self, inplanes, hidden_planes, act = 'ReLU', dp = 0.0, concat = 0):
        super(FC_basic_block, self).__init__()
        self.inplanes = inplanes
        self.hidden_planes = hidden_planes
        self.concat = concat

        self.fc1 = nn.Linear(inplanes, inplanes)
        if act == 'ReLU':
            self.act1 = nn.ReLU(inplace=True)
        self.dp1 = nn.Dropout(dp)

        self.fc2 = nn.Linear(hidden_planes, inplanes)
        if act == 'ReLU':
            self.act2 = nn.ReLU(inplace=True)
        self.dp2 = nn.Dropout(dp)

    def forward(self, x, other_x = None):
        residual = x

        out = self.fc1(x)
        out = self.act1(out)
        out = self.dp1(out)

        if self.inplanes == self.hidden_planes:
            out = self.fc2(out)
        else:
            out = self.fc2(torch.cat([out, other_x], dim=1))

        embeddings = out
        if self.concat == 0:
            out += residual
        else:
            out = torch.cat([out, residual], dim=1)

        out = self.act2(out)
        out = self.dp2(out)

        return out, embeddings


class FC_basic_block_embed(nn.Module):
    def __init__(self, inplanes, hidden_planes, act = 'ReLU', dp = 0.0, concat = 0):
        super(FC_basic_block_embed, self).__init__()
        self.inplanes = inplanes
        self.hidden_planes = hidden_planes
        self.concat = concat

        self.fc1 = nn.Linear(inplanes, inplanes)
        if act == 'ReLU':
            self.act1 = nn.ReLU(inplace=True)
        self.dp1 = nn.Dropout(dp)

        self.fc2 = nn.Linear(hidden_planes, inplanes)
        if act == 'ReLU':
            self.act2 = nn.ReLU(inplace=True)
        self.dp2 = nn.Dropout(dp)

    def forward(self, x, other_x = None):
        residual = x

        out = self.fc1(x)
        out = self.act1(out)
        out = self.dp1(out)

        if self.inplanes == self.hidden_planes:
            out = self.fc2(out)
        else:
            out = self.fc2(torch.cat([out, other_x], dim=1))


        if self.concat == 0:
            out += residual
        else:
            out = torch.cat([out, residual], dim=1)

        out = self.act2(out)
        out = self.dp2(out)

        return out, out


class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'


class GraphConv_cheby(Module):
    def __init__(self, in_features, out_features, K):
        super(GraphConv_cheby, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.K = K
        self.cl = nn.Linear(in_features * K, out_features)
        scale = np.sqrt( 2.0/ (in_features * K + out_features) )
        self.cl.weight.data.uniform_(-scale, scale)
        self.cl.bias.data.fill_(0.0)

    def forward(self, x, adj):
        B, V, Fin = x.size()
        B, V, Fin = int(B), int(V), int(Fin)
        x0 = x.permute(1, 2, 0).contiguous()  # V x Fin x B
        x0 = x0.view([V, Fin * B])  # V x Fin*B
        x = x0.unsqueeze(0)  # 1 x V x Fin*B

        if self.K > 1:
            #print type(x0)
            if type(x0) == torch.cuda.sparse.FloatTensor:
                x1 = my_sparse_mm()(adj, x0)  # V x Fin*B
            else:
                x1 = torch.mm(adj, x0)
            x = torch.cat((x, x1.unsqueeze(0)), 0)  # 2 x V x Fin*B
        for k in range(2, self.K):
            if type(x1) == torch.cuda.sparse.FloatTensor:
                x2 = 2 * my_sparse_mm()(adj, x1) - x0
            else:
                x2 = 2 * torch.mm(adj, x1) - x0
            x = torch.cat((x, x2.unsqueeze(0)), 0)  # M x Fin*B
            x0, x1 = x1, x2

        x = x.view([self.K, V, Fin, B])  # K x V x Fin x B
        x = x.permute(3, 1, 2, 0).contiguous()  # B x V x Fin x K
        x = x.view([B * V, Fin * self.K])  # B*V x Fin*K

        # Compose linearly Fin features to get Fout features
        x = self.cl(x)  # B*V x Fout
        x = x.view([B, V, self.out_features])  # B x V x Fout

        return x

class Graph_pooling(Module):
    def __init__(self, p):
        super(Graph_pooling, self).__init__()
        self.p = p

    def forward(self, x):
        if self.p > 1:
            x = x.permute(0, 2, 1).contiguous()  # x = B x F x V
            x = nn.MaxPool1d(self.p)(x)  # B x F x V/p
            x = x.permute(0, 2, 1).contiguous()  # x = B x V/p x F
            return x
        else:
            return x

class GraphAttentionLayer_MH(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, nheads, dropout, alpha, concat=True):
        super(GraphAttentionLayer_MH, self).__init__()
        self.dropout = dropout
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat

        self.attentions = [GraphAttentionLayer(in_features, out_features, dropout=dropout, alpha=alpha, concat=True) for _ in
                           range(nheads)]

        self.attentions = nn.ModuleList(self.attentions)

    def forward(self, x, adj):
        x = torch.cat([att(x, adj) for att in self.attentions], dim=2)
        x = F.dropout(x, self.dropout, training=self.training)
        return x


class GraphAttentionLayer(nn.Module):
    """
    Simple GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, dropout, alpha, concat=True):
        super(GraphAttentionLayer, self).__init__()
        self.dropout = dropout
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat

        self.W = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_uniform_(self.W.data, gain=1.414)
        self.a = nn.Parameter(torch.zeros(size=(2*out_features, 1)))
        nn.init.xavier_uniform_(self.a.data, gain=1.414)

        self.leakyrelu = nn.LeakyReLU(self.alpha)

    def forward(self, input, adj):
        # input should be B x N x Fin
        B = input.size()[0]
        N = input.size()[1]
        input = input.view(input.size(0) * input.size(1), input.size(2))
        h = torch.mm(input, self.W) # h is B * N x Fout
        h = h.view(B, N, -1)


        a_input = torch.cat([h.repeat(1, 1, N).view(B, N * N, -1), h.repeat(1, N, 1)], dim=1).view(B, N, -1, 2 * self.out_features)
        e = self.leakyrelu(torch.matmul(a_input, self.a).squeeze(3)) # a_input B x N x N * 2Fout

        zero_vec = -9e15*torch.ones_like(e)
        attention = torch.where(adj > 0, e, zero_vec) # attention B x N x N
        attention = F.softmax(attention, dim=2)
        attention = F.dropout(attention, self.dropout, training=self.training)
        h_prime = torch.matmul(attention, h) # h B x N x N
        #print h_prime.size()

        if self.concat:
            return F.elu(h_prime)
        else:
            return h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


class SpecialSpmmFunction(torch.autograd.Function):
    """Special function for only sparse region backpropataion layer."""

    @staticmethod
    def forward(ctx, indices, values, shape, b):
        assert indices.requires_grad == False
        a = torch.sparse_coo_tensor(indices, values, shape)
        ctx.save_for_backward(a, b)
        ctx.N = shape[0]
        return torch.matmul(a, b)

    @staticmethod
    def backward(ctx, grad_output):
        a, b = ctx.saved_tensors
        grad_values = grad_b = None
        if ctx.needs_input_grad[1]:
            grad_a_dense = grad_output.matmul(b.t())
            edge_idx = a._indices()[0, :] * ctx.N + a._indices()[1, :]
            grad_values = grad_a_dense.view(-1)[edge_idx]
        if ctx.needs_input_grad[3]:
            grad_b = a.t().matmul(grad_output)
        return None, grad_values, None, grad_b


class SpecialSpmm(nn.Module):
    def forward(self, indices, values, shape, b):
        return SpecialSpmmFunction.apply(indices, values, shape, b)


class SpGraphAttentionLayer(nn.Module):
    """
    Sparse version GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(self, in_features, out_features, dropout, alpha, concat=True):
        super(SpGraphAttentionLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.concat = concat

        self.W = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        nn.init.xavier_normal_(self.W.data, gain=1.414)

        self.a = nn.Parameter(torch.zeros(size=(1, 2 * out_features)))
        nn.init.xavier_normal_(self.a.data, gain=1.414)

        self.dropout = nn.Dropout(dropout)
        self.leakyrelu = nn.LeakyReLU(self.alpha)
        self.special_spmm = SpecialSpmm()

    def forward(self, input, adj):
        N = input.size()[0]
        edge = adj.nonzero().t()

        h = torch.mm(input, self.W)
        # h: N x out
        assert not torch.isnan(h).any()

        # Self-attention on the nodes - Shared attention mechanism
        edge_h = torch.cat((h[edge[0, :], :], h[edge[1, :], :]), dim=1).t()
        # edge: 2*D x E

        edge_e = torch.exp(-self.leakyrelu(self.a.mm(edge_h).squeeze()))
        assert not torch.isnan(edge_e).any()
        # edge_e: E

        e_rowsum = self.special_spmm(edge, edge_e, torch.Size([N, N]), torch.ones(size=(N, 1)).cuda())
        # e_rowsum: N x 1

        edge_e = self.dropout(edge_e)
        # edge_e: E

        h_prime = self.special_spmm(edge, edge_e, torch.Size([N, N]), h)
        assert not torch.isnan(h_prime).any()
        # h_prime: N x out

        h_prime = h_prime.div(e_rowsum)
        # h_prime: N x out
        assert not torch.isnan(h_prime).any()

        if self.concat:
            # if this layer is not last layer,
            return F.elu(h_prime)
        else:
            # if this layer is last layer,
            return h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


